/*

pbwrelayd.c

TCP daemon/server for relaying of data to remote clients.

Steve Wotton 12-Sep-1994
Based on emurelayd

This is the pbwrelay daemon process. It accepts connections from remote clients
and forks an pbwrelay server process to handle the connection then waits for 
another connection request or for one of the server processes to close the pipe
to which the daemon is forwarding messages.

The -fifo identifies both the TCP port name to which the clients should connect
and the name of the pipe from which pbwrelayd reads the messages. This
helps when trying to determine if one of the pbwrelayd processes are stuck.

*/

#define MAX_MESSAGE_LENGTH 8192
#define DEFAULT_FIFO "pbwrly"
#define DEFAULT_FIFO_DIR "/dev/shm"

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <time.h>
#include <syslog.h>
#include "tcp.h"

/* Not in syslog.h for some reason */

//extern "C" { int syslog(int, const char *, ...); }

int s;				/* connected socket descriptor */
int ls;				/* listen socket descriptor */
int pipefd[2];                  /* Pipe between daemon and server */

struct servent *sp;		/* pointer to service information */

long timevar;			/* contains time returned by time() */
char *ctime();			/* declare time formatting routine */

struct linger linger;		/* allow a lingering, graceful close; */
				/* used when setting SO_LINGER */

struct sockaddr_in myaddr_in;	/* for local socket address */
struct sockaddr_in peeraddr_in;	/* for peer socket address */

static int BroadcastMessage( int, int );
void defunct( int );

int main(int argc, char **argv)
{
int arg, msgfd, i, readfds, Destfds=0;
char *fifo, *fifodir, *fifopath;

   fifo = (char *)malloc( strlen( DEFAULT_FIFO ) + 1 ); strcpy( fifo, DEFAULT_FIFO );

   fifodir = (char *)getenv( "PBW_FIFO_DIR" );
   if ( !fifodir )
   {
     fifodir = (char *)malloc( strlen( DEFAULT_FIFO_DIR ) + 1 ); strcpy( fifodir, DEFAULT_FIFO_DIR );
   }
   
   for ( arg=1;arg<argc;arg++ )
   {
      if ( !strcmp( argv[arg], "-fifo" ) )
      {
         fifo = (char *)realloc( fifo, strlen(argv[arg+1]) + 1 ); strcpy( fifo, argv[arg+1]);
         arg++;
      }
   }
   
   fifopath = (char *)malloc( strlen( fifo ) + strlen( fifodir ) + 2 );
   sprintf( fifopath, "%s/%s", fifodir, fifo );
   while ( (msgfd = open( fifopath, O_RDWR )) == -1){ fprintf(stderr,"errno=%d\n",errno); sleep(5); }

   fcntl( msgfd, F_SETFL, O_NONBLOCK );
   
/* Set the close on exec flag for the msg pipe */

   fcntl( msgfd, F_SETFD, 0x1 );
   
   memset ( (char *)&myaddr_in, 0, sizeof(struct sockaddr_in) );
   memset ( (char *)&peeraddr_in, 0, sizeof(struct sockaddr_in) );

/* Set up address structure for the listen socket. */
		
   myaddr_in.sin_family = AF_INET;
   myaddr_in.sin_addr.s_addr = INADDR_ANY;

   sp = getservbyname( fifo, "tcp" );
   if (sp == NULL)
     myaddr_in.sin_port = htons(7999);
   else
      myaddr_in.sin_port = sp->s_port;

/* Create the listen socket. */
		
   if ( (ls = socket (AF_INET, SOCK_STREAM, 0)) == -1)
   {
      syslog( LOG_ERR | LOG_USER, "pbwrelayd socket(): %m\n" );
      sleep(30);
   }

/* Set the close on exec flag for the listen socket */

   fcntl( ls, F_SETFD, 0x1 );
   
/* Bind the listen address to the socket. */

   while ( bind( ls, (sockaddr *)&myaddr_in, sizeof(struct sockaddr_in) ) == -1 )
   {
      syslog( LOG_ERR | LOG_USER, "pbwrelayd bind(): %m\n" );
      sleep(30);
   }

/* Start listening with up to 5 pending connections allowed */

   while ( listen(ls, 5) == -1 )
   {
      syslog( LOG_ERR | LOG_USER, "pbwrelayd listen(): %m\n" );
      sleep(30);
   }

   setpgrp();

/* Close stderr since its not used and file descriptors are scarce */

   close( 2 );

   switch ( fork() )
   {
   case -1: /* Unable to fork, for some reason. */
      exit(1);
   case 0: /* The daemon */
      signal(SIGCLD, defunct);
      signal(SIGPIPE, SIG_IGN);
      for(;;)
      {
      
/* Wait for new connect,new message or close of one of destinations */

         do
         {
            errno = 0;
            readfds = (1<<ls) | (1<<msgfd);
            select( 32, (fd_set *)&readfds, 0, 0, 0 );
         } while ( errno == EINTR );

/* New message */

         if ( readfds & (1<<msgfd) ) { Destfds = BroadcastMessage( msgfd, Destfds ); continue; }
         else if ( !(readfds & (1<<ls)) ) continue;
         
/* New connect. */

         socklen_t addrlen = (socklen_t)sizeof( struct sockaddr_in );
         s = accept( ls, (sockaddr *)&peeraddr_in, &addrlen );
         if ( s == -1 )
         {
            syslog( LOG_ERR | LOG_USER, "pbwrelayd accept(): %m\n" );
            exit(0);
	 }

         pipe( pipefd );
         if ( pipefd[1] > 31 ) {close(pipefd[0]); close(pipefd[1]); close(s); continue; }

         switch ( fork() )
         {
         case -1: /* Can't fork, just exit. */
            syslog( LOG_ERR | LOG_USER, "pbwrelayd fork(): %m\n" );
            exit(0);
         case 0: /* Child (server) process comes here. */
{
struct hostent *hp;
char *hostname; /* points to the remote host's name string */
unsigned pid;
char c_id[80];

/* Set up file descriptors.  Server will inherit socket as stdout and pipefd[0]
as stdin. pipefd[1] and ls are not used by daemon so they are closed. */

            close( pipefd[1] );
            dup2( s, 1 ); close( s ); s = 1;
            dup2( pipefd[0], 0 ); close( pipefd[0] );

            sockstream S = s;

/* Set the socket linger option */

            linger.l_onoff = 1;
            linger.l_linger = 1;
            S.setopt( SOL_SOCKET, SO_LINGER, &linger, sizeof(linger) );
            if ( S.error )
            {
               syslog( LOG_ERR | LOG_USER, "pbwrelayd setsockopt(): %m\n" );
               exit(1);
	    }

/* Get the host name so that we will be able to see who the server is
connected to */

            hp = gethostbyaddr( (char *)&peeraddr_in.sin_addr, sizeof (struct in_addr), peeraddr_in.sin_family);
            if (hp == NULL) hostname = inet_ntoa(peeraddr_in.sin_addr);
            else hostname = hp->h_name; /* point to host's name */

// Add some info to the argument list of the server so that it appears
// in the ps process list.  The server doesn't actually use any of its
// arguments.

            sprintf( c_id, "%s.%d", hostname, peeraddr_in.sin_port );
            
// Now exec the server process.  Never returns but use exit() after in
// case it fails.

            execlp( "pbwrelay","pbwrelay",c_id,0 );
            syslog( LOG_ERR | LOG_USER, "pbwrelayd execlp(): %m\n" );
            exit(0);
}   
         default: /* Daemon process comes here. */
            close( pipefd[0] );
            close(s);
            Destfds |= 1<<pipefd[1];
            fcntl( pipefd[1], F_SETFD, 0x1 );
         }

      }

      default: /* Parent process comes here. */
         exit(0);
   }
}

static int BroadcastMessage( int msgfd, int Destfds )
{
int i, Length;
char Message[MAX_MESSAGE_LENGTH+1];

/* Get the message from the input FIFO.  Its really not obvious what to do
if there is an error - maybe the following messages will also be corrupt.  At
least ignore the message on error.  */

   for(;;)
   {
      Length = msg_rcv( msgfd, Message, MAX_MESSAGE_LENGTH );
      if ( Length == 0 ) break;
      if ( Length < 0 || Length > MAX_MESSAGE_LENGTH ) return Destfds;

/* The message arrives with newline termination.  Change it to null termination
and adjust length so that it includes terminator.  */

      Message[Length] = '\0';
      Length+=1;
   
/* Try to send the message to all connected servers.  Close down any server that
fails.  */

      for ( i=0;i<32;i++ )
      {
         if ( !(Destfds & ( 1<<i )) ) continue;
         if ( msg_send( i, Message, Length ) < 0) { close(i); Destfds &= ~(1<<i); continue; }
      }
   }
   return Destfds;
}

/* This signal handler is called when a child dies so that a wait() can be
done to remove the <defunct> child process */

void defunct( int dummy )
{
int loc;

   waitpid( -1, &loc, WNOHANG );
   signal( SIGCLD, defunct );
}
