//-*-C++-*-
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <exception>

int msg_send( int , const void *, int );
int msg_rcv( int, void *, int );
int open_tcp_service( const char *, const char * );
int socket_fd( void );

class ss_exception: public std::exception
{
 protected:
  int ecode;
 public:
  ss_exception(int e):exception(),ecode(e){}
  virtual const char* what() const throw()
  {
    return "sockstream exception";
  }
  virtual int errcode() const throw()
  {
    return ecode;
  }
};

class ss_eof: public std::exception
{
 public:
  virtual const char* what() const throw()
  {
    return "sockstream EOF";
  }
};

class sockstream
{
public:
   int s;
   int bufsize;
   int error;
   ~sockstream(){ if ( s != -1 ) close(s); }
   sockstream( void ){ s = socket_fd(); error = s == -1 ? -1 : 0; }
   sockstream( int fd ):s(fd){}
   sockstream( const char *, const char * );
   void setbufsize( int sz ){ bufsize = sz; }
   inline void send( const void *, int );
   inline void rcv( void *, int );
  //   inline void msg_send( const void *, int ); // Deprecated
  //   inline void msg_rcv( void *, int ); // Deprecated
   operator void *(){ return error ? 0 : this; }
   inline sockstream &operator>>( char & );
   inline sockstream &operator>>( unsigned char & );
   inline sockstream &operator<<( char );
   inline sockstream &operator<<( unsigned char );
   inline sockstream &operator>>( unsigned & );
   inline sockstream &operator>>( int & );
   inline sockstream &operator<<( unsigned );
   inline sockstream &operator<<( int );
   inline sockstream &operator>>( char * );
   inline sockstream &operator<<( char * );
   inline sockstream &operator>>( void * );
   inline sockstream &operator<<( void * );
   void setopt( int level, int optname, int val ){ error = setsockopt( s, level, optname, (char *)&val, sizeof(int) ); error = error == -1 ? errno : 0; }
   void setopt( int level, int optname, const void *val, int len ){ error = setsockopt( s, level, optname, val, len ); error = error == -1 ? errno : 0; }
};

inline void sockstream::send( const void *buf, int len )
{
int sent, cc;

  for ( sent=0;sent<len;sent+=cc )
  {
     cc = write( s, (char *)buf + sent, len - sent );
     if( cc == -1 )
     {
       if( errno != EINTR ) throw (new ss_exception(errno));
       cc = 0;
     }
  }
  return;
}

inline void sockstream::rcv( void *buf, int len )
{
int rcvd, cc;

/* Read the message buffer contents */

   for( rcvd=0; rcvd<len; rcvd+=cc )
   {
      cc = read( s, (char*)buf + rcvd, len - rcvd );

      if ( cc == 0 ) throw (new ss_eof()); /* EOF (i.e. socket closed) condition */
      
      if ( cc < 0 )
      {
         if ( errno != EINTR ) throw (new ss_exception(errno));
         cc = 0;
      }
   }
   return;
}

// Get single character from stream

inline sockstream &sockstream::operator>>( char &c )
{
int rc;

   do { rc = read( s, &c, 1 ); } while ( rc == -1 && errno == EINTR );

   if ( rc == -1 ) error = errno;
   else error = rc ? 0 : -1;

   return *this;

}

inline sockstream &sockstream::operator>>( unsigned char &c )
{
   return *this >> (char &)c;
}

// Put single character to stream

inline sockstream &sockstream::operator<<( char c )
{
int rc;

   do { rc = write( s, &c, 1 ); } while ( rc == -1 && errno == EINTR );

   if ( rc == -1 ) error = errno;
   else error = rc ? 0 : -1;

   return *this;

}

inline sockstream &sockstream::operator<<( unsigned char c )
{
   return *this << (char)c;
}

// Get single integer from stream

inline sockstream &sockstream::operator>>( unsigned &i )
{
unsigned char si[4];

   rcv( si, 4 );
   if ( !error )
   {
      i = si[3];
      i |= si[2]<<8;
      i |= si[1]<<16;
      i |= si[0]<<24;
   }

   return *this;

}

inline sockstream &sockstream::operator>>( int &i )
{
   return *this >> (unsigned &)i;
}

// Put single integter to stream

inline sockstream &sockstream::operator<<( unsigned i )
{
unsigned char si[4];

   si[3] = i & 0xff;
   si[2] = (i>>8) & 0xff;
   si[1] = (i>>16) & 0xff;
   si[0] = (i>>24) & 0xff;

   send( si, 4 );

   return *this;

}

inline sockstream &sockstream::operator<<( int i )
{
   return *this << (unsigned)i;
}

// Get string (null terminated character array) from stream including null;

inline sockstream &sockstream::operator>>( char *buf )
{
int nrcv = 0;
int rc;

   while (1)
   {
      do
      {
         rc = read( s, &buf[nrcv], 1 );
      } while ( rc == -1 && errno == EINTR );
      if ( rc < 1 || buf[nrcv] == '\0' ) break;
      nrcv++;
   }

   if ( rc == -1 ) error = errno;
   else error = rc ? 0 : -1;

   return *this;
}

// Put string (null terminated character array) to stream including null;

inline sockstream &sockstream::operator<<( char *buf )
{
int rc, sent;

   int len = strlen( buf ) + 1;

   for ( sent=0;sent<len;sent+=rc )
   {
      rc = write( s, (char *)buf + sent, len - sent );
      if( rc == -1 )
      {
         if( errno != EINTR ) break;
         rc = 0;
      }
   }

   error = rc == -1 ? errno : 0;

   return *this;
}

// Get buffer from stream

inline sockstream &sockstream::operator>>( void *buf )
{

   rcv( buf, bufsize );

   return *this;
}

// Put buffer to stream

inline sockstream &sockstream::operator<<( void *buf )
{
   send( buf, bufsize );

   return *this;
}

// Deprecated (formerly sockstream::send())
// inline void sockstream::send( const void *buf, int len )
// {
//    error = msg_send( s, buf, len );
//    if ( error == -1 ) error = errno;
//    else if ( error != 0 ) error = -1;

//    return;
// }

// Deprecated (formerly sockstream::rcv())
// inline void sockstream::rcv( void *buf, int len )
// {
//    error = msg_rcv( s, buf, len );
//    if ( error == -1 ) error = errno;
//    else if ( error < -1 ) error = -1;
//    else
//    {
//       bufsize = error;
//       error = 0;
//    }

//    return;
// }

